C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE maskline (ityp, regstr, pix, ncol, nrow, mval, ierr)

C @(#)Subroutine: maskline            Author: R.A. Jansen -- Apr 10 2013
C @(#)
C @(#)  Purpose: parse a previously verified region descriptor string in
C @(#)    ASCII pixel list format when its type matches that of a region
C @(#)    encompassing  pixels along a line  with arbitrary  orientation
C @(#)    (ie., 'ityp'=2) or exit otherwise,  and update the input pixel
C @(#)    array by adding the specified mask value if non-zero,  or set-
C @(#)    ting the pixels in the specified pixel region to 0  otherwise.
C @(#)    Any out-of-bounds pixels  that are part of the specified pixel
C @(#)    region will be silently ignored. Should any error condition be
C @(#)    detected, then error status 'ierr' will be non-zero upon exit.
C @(#)  
C @(#)  Calling Sequence:
C @(#)    CALL maskline (ityp, regstr,  pix, ncol, nrow,  mval, ierr)  ,
C @(#)    where 'ityp' is an integer, and 'regstr' is a character string
C @(#)    of arbitrary declared length.   A valid descriptor string con-
C @(#)    sists of four SPACE, TAB, or COMMA-separated numeric values in
C @(#)    the order 'x1 y1 x2 y2', where the line is assumed to run from
C @(#)    (x1,y1) to (x2,y2)  inclusive both endpoints.  'pix' passes an
C @(#)    integer valued array of size 'ncol'x'nrow' pixels (with 'ncol'
C @(#)    and 'nrow' integers).  Both 'mval' and 'ierr' are integers.
C @(#)
C @(#)  Dependencies: several routines from my  'rjgen'  library:  clen,
C @(#)    cinit, creadr, error, parse
C @(#)
C @(#)  Revision history:
C @(#)    Apr 10 2013 [RAJ] -- Initial version of this routine.
C @(#)    Apr 11 2013 [RAJ] -- Modified bounds and x2>x1, y2>y1 checks.
C @(#)    Apr 17 2013 [RAJ] -- Modified bounds check, NINT().
C @(#) -----------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER*(*) regstr
      INTEGER	nrow, ncol, pix(nrow*ncol)
      INTEGER	ityp, mval, ierr

C --- External functions...
      INTEGER	clen

C --- Other program variables...
      CHARACTER delim*1, cpar*80
      REAL	rval, rx1, rx2, ry1, ry2, a, b, ap, bp, xx, yy, dd
      INTEGER	lpar, k, m, n, p, i, j

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Return for a region descriptor string of the wrong type...
      IF (ityp.NE.2) RETURN


C --- Parse the four numeric values from the region descriptor string...
      delim = CHAR(32)
      k     = 0
      m     = clen(regstr)
      DO WHILE (m.GT.0)
         CALL cinit(cpar)
         lpar = 0
         CALL parse(regstr,delim,cpar,lpar)
         IF (lpar.GT.0) THEN
            CALL creadr(cpar(1:lpar),rval,ierr)
         ELSE
            ierr = 1
         ENDIF
         IF (ierr.NE.0) THEN
            CALL error('maskline','parsing LINE descriptor')
            ierr = 1
            RETURN
         ENDIF

         k = k + 1
         IF (k.EQ.1) rx1 = rval
         IF (k.EQ.2) ry1 = rval
         IF (k.EQ.3) rx2 = rval
         IF (k.EQ.4) ry2 = rval

         m = clen(regstr)
      ENDDO

C     Test that we successfully parsed four and only four numeric values
C     from the regions descriptor string...
      IF (k.NE.4) THEN
         CALL cinit(cpar)
         cpar = 'parsing LINE descriptor: nfields != 4'
         CALL error('maskline',cpar)
         ierr = 2
         RETURN
      ENDIF

C     write(*,*)'DEBUG(maskline): read two pairs of x,y coordinates'
C     write(*,*)'DEBUG(maskline):  (x1,y1) = (',rx1,',',ry1,')'
C     write(*,*)'DEBUG(maskline):  (x2,y2) = (',rx2,',',ry2,')'

C     Test whether any pixels are within bounds; an out-of-bounds region
C     or partially out-of-bounds region does not qualify as an error...
      IF (    (NINT(rx1).LT.1.AND.NINT(rx2).LT.1)
     +    .OR.(NINT(ry1).LT.1.AND.NINT(ry2).LT.1)
     +    .OR.(NINT(rx1).GT.ncol.AND.NINT(rx2).GT.ncol)
     +    .OR.(NINT(ry1).GT.nrow.AND.NINT(ry2).GT.nrow)) THEN
C         The specified line segment lies entirely out of bounds; so we
C         have nothing to do...
          RETURN
      ENDIF

C     ------------------------------------------------------------------
C     We use the following method to determine whether the line segment
C     passes through a pixel (i,j) being tested:
C     [0] compute slope and y-intercept for the specified line segment..
C               a = (y2-y1)/(x2-x1) ;  b = y1 - x1*a
C     [1] in case of a finite non-zero slope, compute the y-intercept
C         for a line through the center of pixel (i,j) and perpendicular
C         to the specified line segment...
C               ap = -1/a ;  bp = j - i*ap
C     [2] compute intercept point of the line segment and line through
C         pixel (i,j) perpendicular to that line segment by equating the
C         two line equations...
C               a*x + b = ap*x + bp -->  x = (bp - b)/(a - ap)
C         and hence                      y = x*ap + bp
C     [3] test if the distance from this intercept point to the center
C         of pixel (i,j) is < 0.5*sqrt(2); if so, update pix(j,i) ...
C               d = sqrt( (x-i)**2 + (y-j)**2 )
C
C     Note, that this method is slightly more accurate than just masking
C     the x-intercept with the line segment (which only takes distance
C     along the x-direction into account)
C     ------------------------------------------------------------------

C --- Define the slope and y-intercept of the specified line segment...
      a = (ry2 - ry1)/(rx2 - rx1)
      b = ry1 - rx1*a

C     Define the slope perpendicular to the line segment...
      ap = -(1.0/a)

C     write(*,*)'DEBUG(maskline):  a=',a,', b=',b,', ap=',ap

C     Define maximum distance criterion...
      rval = 0.5*SQRT(2.0)

C --- Update the pixel array if the specified pixel is within bounds...
C     write(*,*)'DEBUG(maskline):  updating pixel array...'
      IF (ABS(rx2-rx1).LT.1.0E-6) THEN
C        Vertical line segment...
         i = NINT(0.5*(rx1+rx2))
         k = MIN(INT(ry1),INT(ry2))
         m = MAX(NINT(ry1),NINT(ry2))
         DO j=MAX(1,k),MIN(nrow,m),1
            IF (mval.EQ.0) THEN
               pix(i+(j-1)*ncol) = 0
            ELSE
               pix(i+(j-1)*ncol) = pix(i+(j-1)*ncol) + mval
            ENDIF
         ENDDO
      ELSE IF (ABS(ry2-ry1).LT.1.0E-6) THEN
C        Horizontal line segment...
         j = NINT(0.5*(ry1+ry2))
         k = MIN(INT(rx1),INT(rx2))
         m = MAX(NINT(rx1),NINT(rx2))
         DO i=MAX(1,k),MIN(ncol,m),1
            IF (mval.EQ.0) THEN
               pix(i+(j-1)*ncol) = 0
            ELSE
               pix(i+(j-1)*ncol) = pix(i+(j-1)*ncol) + mval
            ENDIF
         ENDDO
      ELSE
C        Line segment with finite non-zero slope...
         k = MIN(INT(ry1),INT(ry2))
         m = MAX(NINT(ry1),NINT(ry2))
         n = MIN(INT(rx1),INT(rx2))
         p = MAX(NINT(rx1),NINT(rx2))
         DO j=MAX(1,k),MIN(nrow,m),1
            DO i=MAX(1,n),MIN(ncol,p),1
               bp = j - i*ap
               xx = (bp - b)/(a - ap)
               yy = xx*ap + bp
               dd = SQRT( (xx-i)**2 + (yy-j)**2 )
               IF (dd.LT.rval) THEN
                  IF (mval.EQ.0) THEN
                     pix(i+(j-1)*ncol) = 0
                  ELSE
                     pix(i+(j-1)*ncol) = pix(i+(j-1)*ncol) + mval
                  ENDIF
               ENDIF
            ENDDO
         ENDDO
      ENDIF


      RETURN
      END

