C @(#)[mkmask.f] create/update a mask image,  given one or more lists of
C @(#)           pixel regions (ASCII).      R.A. Jansen -- May  8, 2013
C @(#)
C @(#) Purpose:  create or update a mask image using one or more regions
C @(#)  lists in either DS9 format,  or in one of several old IRAF-style
C @(#)  pixel list formats  (ASCII only),  that may specify rectangular,
C @(#)  circular, elliptical or arbitrary polygonal regions in a corres-
C @(#)  ponding science image. Mask images are flat FITS files, contain-
C @(#)  ing a single header and data unit, with 16-bit int pixel values.
C @(#)  The user may specify specific values to assign unmasked ("good",
C @(#)  default=0) and masked ("bad"; default=1) pixels.   Currently, up
C @(#)  to 8 regions files may be specified at a time.  When updating an
C @(#)  existing mask image, the value of 'goodval' is ignored  (assumed
C @(#)  to have been initialized correctly  when that mask was created);
C @(#)  only new bad pixels will by flagged  by *adding* 'badval' to the
C @(#)  affected pixels in the mask (this behaviour allows binary flags)
C @(#)  with specific meanings (e.g., 1 = bad, 2 = persistent, 4 = satu-
C @(#)  rated, 8 = near edge, etc... and unique combinations thereof).
C @(#)
C @(#) Dependencies: several routines from the CFITSIO[v3.0] I/O library
C @(#)  ('libcfitsio.a'; W.D. Pence), called using their f77 wrappers.
C @(#)
C @(#)  several routines from my 'rjgen' and 'rjfits' libraries, as well
C @(#)  as modified versions of routines in my  'elphot'  and  'imclean'
C @(#)  utilities are included in the [mkmask.f] source file.
C @(#)
C @(#) Compilation instructions:
C @(#)  Check/edit the "makefile" to specify the correct compiler; 'g77'
C @(#)  may be substituted for 'f77' on linux machines.  Also verify the
C @(#)  absolute path in the '-L' directive  to point to the location of
C @(#)  your CFITSIO library if it is not in the standard path. Then:
C @(#)     make
C @(#)  This program compiled under Redhat Enterprise/CentOS5.x linux.
C @(#)
C @(#) Syntax:
C @(#)   (1) When used stand-alone:
C @(#)   mkmask [-h|--help] [-d|--debug [-d]] <mask.fits> \
C @(#)       [ncols=<axlen1>] [nrows=<axlen2>] [bad=<badval>] \
C @(#)       [good=<goodval>] [xoff=<x_offset>] [yoff=<y_offset>] \
C @(#)       <regions_list1> [[regions_list2] ...]
C @(#)
C @(#)   (2) When used within IRAF (using scripted task "mkmask.cl"):
C @(#)   mkmask ("mask.fits", ncols, nrows, "space_separated_file_list",
C @(#)       goodval=0, badval=1, xoff=0., yoff=0., verbose=yes)
C @(#)
C @(#)   NB: the 'ncols' and 'nrows' input parameters are mandatory when
C @(#)       creating a new mask image,  but are ignored (if given) when
C @(#)       updating  an existing mask.  Optional parameters  'goodval'
C @(#)       and  'badval'  will be cast to the nearest integer when not
C @(#)       given as integer values.
C @(#)
C @(#) Example:
C @(#)   (1) $>  mkmask smc136i5_mask.fits ncols=4176 nrows=2180 bad=1 \
C @(#)       good=0 smc136i5brstarph.reg smc136i5bgobjph.reg
C @(#)   (2) cl>  mkmask ("smc136i5_mask.fits", 4176, 2180,
C @(#)       "smc136i5brstarph.reg smc136i5bgobjph.reg", badval=1,
C @(#)       goodval=0, verbose=yes)
C @(#)
C @(#) Revision history:
C @(#)  Jan 14 2013 [TSH/RAJ] -- Theoretical outline of polygon masking.
C @(#)  Jan 15 2013 [RAJ] -- Extracted relevant I/O routines and parsing
C @(#)                       subroutines from 'elphot.f' (v June 2005).
C @(#)  Jan 17 2013       -- Designed and wrote the routine to parse all
C @(#)  -- Feb 15 2013 [RAJ]     command line input; wrote manual page.
C @(#)  Feb 15 2013 [RAJ] -- Wrote script to test routine 'clinput'.
C @(#)  Mar 11 2013       -- Wrote 'makefile'; wrote FITS I/O, regions
C @(#)  -- Apr 24 2013 [RAJ]     descriptor parsing, and ellipse & poly-
C @(#)                           gon masking routines.
C @(#)  May  8 2013 [RAJ] -- Updated documentation and calling sequences
C @(#)                       of 'clinput' and 'regtyp' to include 'xoff'
C @(#)                       and 'yoff' parameters. Updated manual page.
C @(#)  May 13 2013 [RAJ] -- Throughout pix(j,i) -> pix(i+(j-1)*ncol)
C @(#) -----------------------------------------------------------------

      PROGRAM mkmask

      IMPLICIT NONE

C --- Learn about default devices...
      INCLUDE	'devices.inc'

C --- Learn about maximum image dimension, and maximum axis lengths...
      INCLUDE	'buffers.inc'

C --- Declaration of program variables...
C     -- Integer error status, verbosity level; routine identifier...
      INTEGER	ierr, idbg
      CHARACTER	dbid*19

C     -- File I/O variables and boolean switch...
      CHARACTER*80 mask, files(8)
      INTEGER	mlen, flen(8), nfil
      LOGICAL	update

C     -- Memory I/O variables...
      CHARACTER*80 sstr, ctmp, cline*2048
      INTEGER	axlen(MAXAX), pix(MAXSIZ2*MAXSIZ1)
      INTEGER	slen, ilen, ncol, nrow

C     -- Other program variables and loop counters...
      REAL	xoff, yoff
      INTEGER	goodval, badval, ityp, i, j
      LOGICAL	eof


C --------------------------   M  A  I  N   ----------------------------


C --- Initialize any variables not initialized on exit of 'clinput'...
      ierr   = 0
      idbg   = 0
      dbid   = 'DEBUG(mkmask):     '


C --- Read, parse, and verify the command-line input:  if succesful,  we
C     will have read the name of a mask image (which may exist already),
C     the name of  one or more regions files,  the dimension of the mask
C     image, the values denoting good and masked pixels, offsets between
C     regions file and image coordinates, and a verbosity modifier...
      CALL debugmsg(1,dbid,idbg,' ',0,0.)
      CALL clinput (mask, mlen, update, files, flen, nfil, axlen,
     +              goodval, badval, xoff, yoff, idbg, ierr)
      IF (ierr.EQ.-1) CALL manual()
      IF (ierr.NE.0) GOTO 10


C     Prepare first part of opening message...
      CALL wsizstr (2, 2, axlen, sstr, slen, ierr)
      IF (ierr.NE.0) GOTO 9
      ilen = 8 + mlen + slen
      ctmp(1:ilen) = 'MKMASK: '//mask(1:mlen)//sstr(1:slen)

C --- Are we updating an existing mask image or creating a new one?
      IF (update) THEN
C        Write opening message for existing mask image...
         WRITE(STDOUT,'(A)') ctmp(1:ilen)//'   [update=yes]'

C        Read the input mask image into integer pixel array pix(j,i)...
         CALL rcfitsi (mask, pix, ncol, nrow, idbg, ierr)
         IF (ierr.NE.0) CALL error('mkmask','reading input mask image')
         IF (ierr.NE.0) GOTO 9

         IF (ncol.NE.axlen(1).OR.nrow.NE.axlen(2)) THEN
            ctmp = 'axis lengths reported by "rcfitsi" do not match '//
     +             'those reported by "clinput".'
            CALL error('mkmask',ctmp)
            GOTO 9
         ENDIF

CC******* test
CC        write out FITS mask image just read in...
C         CALL wcfitsi ('asread.fits', pix, ncol, nrow, .FALSE.,
C     +              idbg, ierr)
C         IF (ierr.NE.0) CALL error('mkmask','writing asread image')
C         IF (ierr.NE.0) GOTO 9
CC******* end-test

      ELSE
C        Write opening message for new mask image...
         WRITE(STDOUT,'(A)') ctmp(1:ilen)//'   [update=no]'

C        Initialize pixel array 'pix' to 'goodval' (empty mask)...
         ncol = axlen(1)
         nrow = axlen(2)
         DO j=1,nrow
            DO i=1,ncol
               pix(i+(j-1)*ncol) = goodval
            ENDDO
         ENDDO

CC******* test
CC        write out image buffer just created as a FITS file...
C         CALL wcfitsi ('asinit.fits', pix, ncol, nrow, .FALSE.,
C     +              idbg, ierr)
C         IF (ierr.NE.0) CALL error('mkmask','writing asinit image')
C         IF (ierr.NE.0) GOTO 9
CC******* end-test


      ENDIF


C --- Read and parse each of the regions files in turn...
      DO i=1,nfil

C        Open the regions file, then read and try to parse each line...
         CALL opendev(STDIFIL,files(i),'read','ascii',0,ierr)
         IF (ierr.NE.0) GOTO 9

C        Loop over the lines in the current regions file...
         eof = .FALSE. 

C        Try and parse the next line from the current regions file...
1        CALL readline(STDIFIL, cline, ilen, eof, ierr)
         CALL debugmsg(2,dbid,idbg,cline,0,0.)
         IF (ierr.NE.0) GOTO 9
         IF (eof) GOTO 2

C        Try and identify the type of region descriptor  (-1=unknown,
C        0=comment, 1=point/pixel, 2=line segment, 3=rectangular box,
C        4=ellipse (incl. circle), and 5=polygon (incl. rotated box),
C        possibly translating region descriptors to one of five basic
C        types in the simpler ASCII pixel list format and applying any
C        specified offsets in x and y  to the relevant regions coords
C        (in which case 'cline' will be returned modified upon exit)..
         CALL regtyp (cline, ilen, ityp, xoff, yoff, idbg, ierr)

         IF (ierr.NE.0) THEN
            CALL cinit(ctmp)
            ctmp(1:39) = 'interpreting region descriptor string "'
            IF (ilen.LE.40) THEN
               ctmp(40:(40+ilen)) = cline(1:ilen)
               ctmp(41+ilen:41+ilen) = '"'
            ELSE
               ctmp(40:77) = cline(1:37)
               ctmp(78:80) = '.."'
            ENDIF
            CALL error('mkmask',ctmp)
            GOTO 9
         ENDIF

         IF (ityp.EQ.-1) THEN
            ctmp(1:47)='skipping unsupported region descriptor string "'
            IF (ilen.LE.32) THEN
               ctmp(48:(48+ilen)) = cline(1:ilen)
               ctmp(49+ilen:49+ilen) = '"'
            ELSE
               ctmp(48:77) = cline(1:29)
               ctmp(78:80) = '.."'
            ENDIF
            CALL error('mkmask',ctmp)
            ierr = 0

         ELSE IF (ityp.EQ.0) THEN
C           Successfully identified a line we may skip...
            CONTINUE

         ELSE
C           Successfully identified a region descriptor...
            CALL debugmsg(3,dbid,idbg,' ',ityp,(1.0*ilen))

C           Update buffer 'pix' by adding 'badval' to each pixel that is
C           encompassed by  the current region.  (the following routines
C           only operate on region descriptors that match their specific
C           'ityp' and immediately return otherwise) ...
            CALL maskpoint (ityp,cline,pix,ncol,nrow,badval,ierr)
            CALL maskline (ityp,cline,pix,ncol,nrow,badval,ierr)
            CALL maskrectang (ityp,cline,pix,ncol,nrow,badval,ierr)
            CALL maskellipse (ityp,cline,pix,ncol,nrow,badval,ierr)
            CALL maskpolygon (ityp,cline,pix,ncol,nrow,badval,ierr)
            IF (ierr.NE.0) GOTO 9

         ENDIF

C        Continue with the next line if we haven't reached EOF yet...
         IF (.NOT.eof) GOTO 1


C        Status report...
2        CALL cinit(ctmp)
         ilen = flen(i) + 18
         IF (update.OR.i.GT.1) THEN
            ctmp(1:ilen) = '  added "'//files(i)(1:flen(i))//'" to mask'
         ELSE
            ctmp(1:ilen) = '  converted "'//files(i)(1:flen(i))//'"    '
         ENDIF
         WRITE(STDOUT,'(A)') ctmp(1:ilen)

C        Close the regions file, then continue to next regions file...
         CALL debugmsg(4,dbid,idbg,files(i),0,0.)
         CLOSE (STDIFIL)

      ENDDO


C --- Write out new or updated FITS mask image...
      CALL wcfitsi (mask, pix, ncol, nrow, update, idbg, ierr)
      IF (ierr.NE.0) CALL error('mkmask','writing mask image')
      IF (ierr.NE.0) GOTO 9


C --- End message in case of success...
      CALL debugmsg(5,dbid,idbg,' ',0,0.)
      STOP

C --- End message in case of any error condition...
9     WRITE(STDOUT,'(A)')'MKMASK: Aborted.'

10    STOP
      END


