C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE regtyp (regstr, ilen, ityp, xoff, yoff, idbg, ierr)

C @(#)Subroutine: regtyp              Author: R.A. Jansen -- Apr 24 2013
C @(#)
C @(#)  Purpose:  Parse a region descriptor string and identify the type
C @(#)    of region being encoded.  If recognized, the descriptor string
C @(#)    will be translated into one of five basic types in the simpler
C @(#)    ASCII pixel list format and returned through 'regstr' on exit,
C @(#)    and the actual length of 'regstr'  will be returned by 'ilen'.
C @(#)    If a comment line, 'regstr' will be blank  and both 'ilen' and
C @(#)    'ityp' will be 0 on exit.
C @(#)    On input, descriptors in various formats --such as DS9 regions
C @(#)    format, and various ASCII pixel list formats-- are acceptable.
C @(#)    The returned value of 'ityp' has the following meaning:
C @(#)
C @(#)     ityp    interpretation          output pixel list format
C @(#)     -1  unrecognized descriptor     --
C @(#)      0  (treat as) a comment line   --
C @(#)      1  single point/pixel          x0 y0
C @(#)      2  line segment                x1 y1 x2 y2
C @(#)      3  rectangular box (no rot.)   x1 x2 y1 y2
C @(#)      4  ellipse (incl. circle)      x0 y0 da db PA
C @(#)      5  polygon (incl. rotated box) x1 y1 x2 y2 .. xN yN for N>=3
C @(#)
C @(#)    If verbosity level 'idbg'>0,  additional messages are printed.
C @(#)    If any error condition is detected,  error status 'ierr'  will
C @(#)    be non-zero upon exit.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL regtyp (regstr, ilen, ityp, xoff, yoff, idbg, ierr)     ,
C @(#)    where  'regstr'  is a character string  of arbitrary  declared
C @(#)    length,  integer 'ilen'  returns the actual length of 'regstr'
C @(#)    upon exit after possible translation,  integer 'ityp'  returns
C @(#)    a value between -1 and 5  corresponding to the identified type
C @(#)    of region encoded.  Real values 'xoff' and 'yoff'  are offsets
C @(#)    to apply to coordinates read from the region descriptor.  Both
C @(#)    'idbg' and 'ierr' are integers.  All offsets are in the sense:
C @(#)    x_mask = x_reg + xoff  and  y_mask = y_reg + yoff .
C @(#)
C @(#)  Dependencies: nparse; several routines from my  'rjgen' library:
C @(#)    cidx, cinit, clen, creadr, error, parse, ucase.
C @(#)
C @(#)  Note on the inner workings -- we filter out and/or test for:
C @(#)    "# Region file format: DS9" -> ok; treat as comment? or should
C @(#)       we add 10 to all 'ityp' values and enforce the remainder of
C @(#)       regions descriptors to strictly adhere?
C @(#)    "#"/"physical"/"image"/"global"/"local"/"text"/"ruler" -> ok,
C @(#)       treat as if they are comments
C @(#)    "fk4"/"fk5"/"B1950"/"J2000"/"ICRS" -> not supported (yet)
C @(#)    "  " (blank lines) -> ok; treat as comment (or issue warning?)
C @(#)    "point()"/"box()"/"line()"/"circle()"/"ellipse()"/"polygon()"
C @(#)    "x0 y0" / "x0 y0 r" / "x1 x2 y1 y2" / "x1 x2 da db PA" / and
C @(#)    "x1 y1 x2 y2 .. xN yN" for all N>=3.
C @(#)
C @(#)  Note, that this routine only supports a subset of the full range
C @(#)    of allowed DS9 regions format descriptors, and requires use of
C @(#)    parentheses and allows only one descriptor per line (i.e., ';'
C @(#)    is not supported as a separator of descriptors).
C @(#)
C @(#)  Revision history:
C @(#)    Apr 24 2013 [RAJ] -- Initial 'working' version of this routine
C @(#)    Apr 30-May 7[RAJ] -- Extensive debugging...
C @(#)    May  8 2013 [RAJ] -- Added 'xoff' & 'yoff' to calling sequence
C @(#)                         to allow offsets of regions coordinates
C @(#) -----------------------------------------------------------------

      IMPLICIT NONE

C --- Learn about default devices...
      INCLUDE	'devices.inc'

C --- Declaration of input/output variables...
      CHARACTER*(*) regstr
      REAL	xoff, yoff
      INTEGER	ilen, ityp, idbg, ierr

C --- External functions...
      INTEGER	cidx, clen
      REAL	SIND, COSD, ATAN2D
      EXTERNAL	SIND, COSD, ATAN2D

C --- Other program variables...
      CHARACTER	ctmp*2048, cstr*2048, cmsg*80, dbid*19
      REAL	pars(512), x0, y0, da, db, PA, rad, phi1, phi2
      INTEGER	slen, i, j, k, m, n
      LOGICAL	lint, ds9

C     ------------------------------------------------------------------


C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Set subroutine/function identifier...
      dbid = 'DEBUG(regtyp):     '

C --- Initialize output and local variables...
      ilen = -1
      ityp = -1

C --- Convert the supplied descriptor string to upper case...
      CALL ucase(regstr,ctmp)
      slen = clen(ctmp)

C     For highly raised verbosity levels, echo the supplied descriptor
C     and capture the special case of a header line announcing a DS9
C     regions file format...
      CALL debugmsg(51,dbid,idbg,regstr(1:slen),0,0.)
      IF (ctmp(1:25).EQ.'# REGION FILE FORMAT: DS9') THEN
         CALL debugmsg(52,dbid,idbg,' ',0,0.)
      ENDIF

C --- Check whether the region descriptor string is, or should be
C     treated as a comment line...
      IF (slen.EQ.0) THEN
C        Detected a blank line on input; ignore, but warn...
         CALL debugmsg(53,dbid,idbg,' ',0,0.)
         CALL cinit(regstr)
         ilen = 0
         ityp = 0
         RETURN
      ENDIF

      IF (    ctmp(1:1).EQ.'#'
     +    .OR.ctmp(1:5).EQ.'LOCAL'
     +    .OR.ctmp(1:6).EQ.'GLOBAL'
     +    .OR.ctmp(1:5).EQ.'IMAGE'
     +    .OR.ctmp(1:8).EQ.'PHYSICAL'
     +    .OR.ctmp(1:8).EQ.'DETECTOR'
     +    .OR.ctmp(1:9).EQ.'AMPLIFIER'
     +    .OR.ctmp(1:4).EQ.'TEXT'
     +    .OR.ctmp(1:5).EQ.'RULER'    ) THEN
C        Detected a comment line or a line we may treat as one...
         CALL debugmsg(54,dbid,idbg,' ',0,0.)
         CALL cinit(regstr)
         ilen = 0
         ityp = 0
         RETURN
      ENDIF

      IF (    ctmp(1:3).EQ.'FK4'
     +    .OR.ctmp(1:3).EQ.'FK5'
     +    .OR.ctmp(1:4).EQ.'ICRS'
     +    .OR.ctmp(1:5).EQ.'B1950'
     +    .OR.ctmp(1:5).EQ.'J2000'
     +    .OR.ctmp(1:6).EQ.'LINEAR'
     +    .OR.ctmp(1:8).EQ.'GALACTIC'
     +    .OR.ctmp(1:8).EQ.'ECLIPTIC') THEN
C        Detected a world coordinate specifier; this routine at present
C        only interprets regions specified in pixel coordinates...
         CALL cinit(cmsg)
         cmsg = 'world coordinate descriptors are not supported'
         CALL error('regtyp',cmsg)
         ilen = slen
         ityp = -1
         ierr = 1
         RETURN
      ENDIF


C --- We appear to have read a line that describes a region; try to
C     identify its type, starting with DS9 style descriptors...

C     Search for the (in this routine) mandatory opening parenthesis
C     of a DS9-style descriptor...
      k = cidx(ctmp,'(')
      k = k - 1
      IF (k.GE.3.AND.(    ctmp(1:k).EQ.'POINT'
     +                .OR.ctmp(1:k).EQ.'BOX'
     +                .OR.ctmp(1:k).EQ.'LINE'
     +                .OR.ctmp(1:k).EQ.'CIRCLE'
     +                .OR.ctmp(1:k).EQ.'ELLIPSE'
     +                .OR.ctmp(1:k).EQ.'POLYGON')) THEN

C        Looks like a supported DS9-style region descriptor string...
         ds9 = .TRUE.

         CALL debugmsg(55,dbid,idbg,ctmp(1:slen),0,0.)

C        -- Some basic format tests...
C        If here, we know the position of the opening parenthesis...
         i = k + 2

C        Search for the (first) closing parenthesis...
         j = cidx(ctmp,')')
         IF (j.EQ.0) THEN
            CALL cinit(cmsg)
            cmsg = 'region descriptor misses closing parenthesis'
            CALL error('regtyp',cmsg)
            ilen = slen
            ityp = -1
            ierr = 2
         ENDIF
         j = j - 1

C        While the DS9 region format allows multiple descriptors per
C        line separated by a semi-colon, the present routine enforces
C        a limit of only one descriptor per line (lazy programmer)...
         m = cidx(ctmp,';')
         n = cidx(ctmp,'#')
         IF (m.NE.0.AND.(n.EQ.0.OR.m.LT.n)) THEN
            CALL cinit(cmsg)
            cmsg = '[;] multiple descriptors per line not supported'
            CALL error('regtyp',cmsg)
            ilen = slen
            ityp = -1
            ierr = 2
         ENDIF

         IF (ierr.NE.0) RETURN

C        Excise relevant portion from the region descriptor string...
         CALL cinit(cstr)
         m = j - i + 1
         cstr(1:m) = ctmp(i:j)
         CALL debugmsg(56,dbid,idbg,cstr(1:m),0,0.)

      ELSE

C        Not a DS9-style descriptor; test ASCII pixel list format...
         ds9 = .FALSE.

         CALL cinit(cstr)
         cstr(1:slen) = ctmp(1:slen)
         CALL debugmsg(57,dbid,idbg,cstr(1:slen),0,0.)

C        Parse and count the number of fields: 2-->point, 3-->circle,
C        4-->rectangle (no rot.), 5-->ellipse, 2*N-->polygon
         n    = 512   
         lint = .FALSE.
         idbg = -1 - idbg
         CALL nparse (cstr, n, pars, lint, idbg, ierr)
         IF (ierr.NE.0) THEN
            CALL error('regtyp','nparse returned with ierr>0')
            ierr = 10
            RETURN
         ENDIF
         CALL debugmsg(58,dbid,idbg,' ',n,0.)

C        Restore 'cstr' (it was returned empty after 'nparse' call)...
         cstr(1:slen) = ctmp(1:slen)

      ENDIF


C     Test each of the supported regions descriptors in turn...
      IF (ds9) THEN
C        Apparent DS9-style region descriptor...
         CALL debugmsg(59,dbid,idbg,' ',0,0.)

         IF (ctmp(1:k).EQ.'POINT') THEN
C           Looks like a single pixel region descriptor of the form:
C           point(x,y) ...
            n    = 2
            lint = .TRUE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 3
               RETURN
            ENDIF

C           Definitely a single pixel region descriptor; translate into
C           a simpler ASCII pixel list format...
            ityp = 1 
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(1)+xoff),(pars(2)+yoff)

C           write(*,*) dbid//'  n=',n,'  ',(pars(1)+xoff),(pars(2)+yoff)

         ELSE IF (ctmp(1:k).EQ.'LINE') THEN
C           Looks like a line segment region descriptor of the form:
C           line(x1,y1,x2,y2) ...
            n    = 4
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 4
               RETURN
            ENDIF

C           Definitely a line segment region descriptor; translate into
C           a simpler ASCII pixel list format...
            ityp = 2
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + yoff
            pars(3) = pars(3) + xoff
            pars(4) = pars(4) + yoff
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n)

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (ctmp(1:k).EQ.'BOX') THEN
C           Looks like a rectangular region descriptor of the form:
C           box(x0,y0,da,db,PA) --> treat as "rectangle" when PA=0deg
C           or PA=+/-90deg and as "polygon" otherwise ...
            n    = 5
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 5
               RETURN
            ENDIF

C           Definitely a box region descriptor; test whether it is a
C           rotated box or one aligned with the pixel grid...
            x0 = pars(1) + xoff
            y0 = pars(2) + yoff
            da = pars(3)
            db = pars(4)
            PA = pars(5)
            IF (ABS(MOD(PA,90.0)).LT.1.0E-5) THEN
C              Simple rectangular box, aligned with pixel grid...
               ityp = 3
               CALL cinit(cstr)
               IF (MOD(NINT(PA),180).EQ.0) THEN
C                 Effective orientation is 0(=180=360) degrees... 
                  WRITE(cstr,*) NINT(x0-0.5*da),NINT(x0+0.5*da),
     +                          NINT(y0-0.5*db),NINT(y0+0.5*db)
               ELSE
C                 Effective orientation is -90(=90=270) degrees...
                  WRITE(cstr,*) NINT(x0-0.5*db),NINT(x0+0.5*db),
     +                          NINT(y0-0.5*da),NINT(y0+0.5*da)
               ENDIF
            ELSE
C              Rectangular box, rotated wrt. the pixel grid; we will
C              treat this region as a polygon instead, so we need to
C              compute each of the vertices...
               ityp = 5
               CALL cinit(cstr)
               rad = 0.5*SQRT(da**2 + db**2)
               phi1 = ATAN2D(db,da) + PA
               phi2 = ATAN2D(db,da) - PA
               WRITE(cstr,*) (x0-rad*COSD(phi1)),(y0-rad*SIND(phi1)),
     +                       (x0+rad*COSD(phi2)),(y0-rad*SIND(phi2)),
     +                       (x0+rad*COSD(phi1)),(y0+rad*SIND(phi1)),
     +                       (x0-rad*COSD(phi2)),(y0+rad*SIND(phi2))
            ENDIF

C           write(*,*) dbid//'  cstr="'//cstr(1:clen(cstr))//'"'

         ELSE IF (ctmp(1:k).EQ.'CIRCLE') THEN
C           Looks like a circular region descriptor of the form:
C           circle(x0,y0,r) --> translate to "ellipse" with PA=0
            n    = 3
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 6
               RETURN
            ENDIF

C           Definitely a circular region descriptor; translate into a
C           simpler ASCII pixel list format...
            ityp = 4
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + yoff
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n), pars(n), 0.0

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (ctmp(1:k).EQ.'ELLIPSE') THEN
C           Looks like an elliptical region descriptor of the form:
C           ellipse(x0,y0,da,db,PA) ...
            n    = 5
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 7
               RETURN
            ENDIF

C           Definitely an elliptical region descriptor; translate into a
C           simpler ASCII pixel list format...
            ityp = 4
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + yoff
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n)

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (ctmp(1:k).EQ.'POLYGON') THEN
C           Looks like an arbitrary polygon region descriptor of the
C           form:  polygon(x1,y1,x2,y2,...,xN,yN) for 3<=N<=256.  We
C           won't know N a priori and so must call 'nparse' with idbg<0
C           to suppress an error message when failing to read 512 values
C           (256 x,y pairs) and report the number read instead (idbg
C           will be reset to its original >=0 value upon exit) ...
            CALL cinit(cmsg)
            n    = 512
            lint = .FALSE.
            idbg = -1 - idbg
            IF (idbg.LT.-2) THEN
               WRITE(STDOUT,*) dbid//' clen(cstr)=',clen(cstr),
     +               ', n=',n,', idbg=',idbg
            ENDIF

C           write(*,*)dbid//'[before nparse] cstr="'//cstr(1:80)//'"'
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
C           write(*,*)dbid//'[after nparse] cstr="'//cstr(1:80)//'"'
            IF (ierr.NE.0) THEN
               CALL error('regtyp','nparse returned with ierr>0')
               ierr = 8
               RETURN
            ENDIF

            IF (n.LT.6) THEN
C              Polygon must have N>=3 vertices (n>=6) ...
               cmsg = 'number of vertices less than minimum (N<3)'
               CALL error('regtyp',cmsg)
               ierr = 8
            ENDIF
            IF (n.GT.512) THEN
C              Internal buffer size in 'maskpolygon' exceeded...
               cmsg = 'number of vertices exceeds maximum (N>256)'
               CALL error('regtyp',cmsg)
               ierr = 8
            ENDIF
            IF (MOD(n,2).NE.0) THEN
C              Odd number of values; (x,y) pairs must be even...
               cmsg = 'odd number of values parsed from descriptor'
               CALL error('regtyp',cmsg)
               ierr = 8
            ENDIF
            IF (ierr.NE.0) RETURN

C           If here, definitely a polygon region descriptor; copy to a
C           simpler ASCII pixel list format...
            ityp = 5
            DO m=1,n
               IF (MOD(m,2).EQ.1) THEN
                  pars(m) = pars(m) + xoff
               ELSE
                  pars(m) = pars(m) + yoff
               ENDIF
            ENDDO
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n)

         ELSE
C           It should be impossible to get here...
            CALL error('regtyp','failed impossibly')
            WRITE(STDOUT,'(A)') dbid//'last descriptor read "'//
     +            ctmp(1:k)//'()"'
            ierr = 99
            RETURN

         ENDIF

         CALL debugmsg(60,dbid,idbg,ctmp(1:k),0,0.)

      ELSE
C        Not a DS9-style region desriptor...
         CALL debugmsg(61,dbid,idbg,' ',0,0.)

C        write(*,*) dbid//' idbg=',idbg,' n=',n,', cstr="'//cstr//'"'

         CALL cinit(cmsg)
         IF (n.LT.2) THEN
            ilen = slen
            cmsg = 'failed to read a minimum of two fields (n<2)'
            CALL error('regtyp',cmsg)
            ierr = 11
            RETURN

         ELSE IF (n.EQ.2) THEN
C           Looks like a single pixel region descriptor...
            lint = .TRUE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 12
               RETURN
            ENDIF

            ityp = 1
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(1)+xoff),(pars(2)+yoff)

C           write(*,*) dbid//'  n=',n,'  ',(pars(1)+xoff),(pars(2)+yoff)

         ELSE IF (n.EQ.3) THEN
C           Looks like a circular pixel region descriptor...
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 13
               RETURN
            ENDIF

            ityp = 4
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + yoff
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n), pars(n), 0.0

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (n.EQ.4) THEN
C           Looks like a descriptor for a simple rectangular pixel
C           region, aligned with pixel grid (x1 x2 y1 y2), or for a
C           single pixel (x1 x1 y1 y1 --> x1 y1) ...
            lint = .TRUE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 14
               RETURN
            ENDIF
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + xoff
            pars(3) = pars(3) + yoff
            pars(4) = pars(4) + yoff
            CALL cinit(cstr)
            IF (pars(1).EQ.pars(2).AND.pars(3).EQ.pars(4)) THEN
               ityp = 1
               WRITE(cstr,*) pars(1),pars(3)

C              write(*,*) dbid//'  n=2  ',pars(1),pars(3)
            ELSE
               ityp = 3
               WRITE(cstr,*) (pars(m),m=1,n)

C              write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)
            ENDIF

        ELSE IF (n.EQ.5) THEN
C           Looks like an elliptical pixel region (x0 y0 da db PA)...
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 15
               RETURN
            ENDIF

            ityp = 4
            pars(1) = pars(1) + xoff
            pars(2) = pars(2) + yoff
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n)

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (n.GE.5.AND.n.LE.512.AND.MOD(n,2).EQ.0) THEN
C           Looks like a polygon (x1 y1 x2 y2 ... xN yN) for 3<=N<=256..
            lint = .FALSE.
            CALL nparse (cstr, n, pars, lint, idbg, ierr)
            IF (ierr.NE.0) THEN
               ierr = 16
               RETURN
            ENDIF

            ityp = 5
            DO m=1,n
               IF (MOD(m,2).EQ.1) THEN
                  pars(m) = pars(m) + xoff
               ELSE
                  pars(m) = pars(m) + yoff
               ENDIF
            ENDDO
            CALL cinit(cstr)
            WRITE(cstr,*) (pars(m),m=1,n)

C           write(*,*) dbid//'  n=',n,'  ',(pars(m),m=1,n)

         ELSE IF (n.GT.512) THEN
C           Internal buffer size in 'maskpolygon' exceeded...
            cmsg = 'number of vertices exceeds maximum (N>256)'
            CALL error('regtyp',cmsg)
            ilen = slen
            ierr = 17
            RETURN

         ELSE IF (MOD(n,2).NE.0) THEN
C           Odd number of values; (x,y) pairs must be even...
            cmsg = 'odd number of values parsed from descriptor'
            CALL error('regtyp',cmsg)
            ierr = 18
            RETURN

         ENDIF
            
      ENDIF


C     Return possibly modified/simplified regions string and its actual
C     length...
      CALL cinit(regstr)
      ilen = clen(cstr)
      IF (ilen.GT.0) regstr(1:ilen) = cstr(1:ilen)
      CALL debugmsg(62,dbid,idbg,cstr,ilen,0.)
 

9     RETURN
      END

