C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++   

      SUBROUTINE rjfchdr (iext, fhdr, nrec, nfree, idbg, ierr)

C @(#)Subroutine: rjfchdr            Author: R.A. Jansen -- Jul  8, 2009
C @(#)
C @(#)  Purpose:  Create a minimally populated FITS header array  'fhdr'
C @(#)    in memory for either a simple flat FITS file ('iext'=-1)  or a
C @(#)    specified HDU within a multi-extension FITS file ('iext' >= 0,
C @(#)    where 0 denotes the Primary Header Unit, 1 the first Extension
C @(#)    Header Unit, etc..).  In case of a flat FITS file, 'iext' will
C @(#)    reset internally to 0 and retain that value on exit.  The num-
C @(#)    of header records written is returned thru element 'iext+1' of
C @(#)    integer array  'nrec(MAXEXT)'  and will be an integer multiple
C @(#)    of 36 (i.e., a 2880-byte logical header block).  The number of
C @(#)    free records in the final block of the present HDU  (following
C @(#)    the "END" record) is returned thru element 'iext+1' of integer
C @(#)    array 'nfree(MAXEXT)'.  The new header array  is returned thru
C @(#)    elements  'nrec(iext)+1' through 'nrec(iext)+nrec(iext+1)'  of
C @(#)    character array 'fhdr(MAXREC)',  where 'nrec(0)' would need to
C @(#)    be evaluated as 1.  If integer verbosity level 'idbg'>0, addi-
C @(#)    tional messages may be printed.   Integer error status  'ierr'
C @(#)    will be non-zero upon exit if any error condition is detected.
C @(#)    Its value (if >2000) uniquely identifies a known condition.
C @(#)
C @(#)  Calling sequence:  CALL rjfchdr(iext,fhdr,nrec,nfree,idbg,ierr),
C @(#)    where 'iext' is an integer  with a valid range of [-1..MAXEXT]
C @(#)    upon invokation and of [0..MAXEXT] on exit, 'fhdr' is a MAXREC
C @(#)    element character array  of 80-char elements,  and  'nrec' and
C @(#)    'nfree'  are MAXEXT element integer arrays.  'idbg' and 'ierr'
C @(#)    are both integers.
C @(#)
C @(#)  Note:  The header array 'fhdr' has no special internal structure
C @(#)    to delimit records belonging to different header units. Appen-
C @(#)    ding a record may require shifting existing records  of either
C @(#)    the current or of subsequent header units by 1 or 36 elements.
C @(#)    The external integer arrays  'nrec' and 'nfree'  keep track of
C @(#)    the intended logical structure of the FITS header.   They must
C @(#)    not be re-initialized or modified in the calling routine,  but
C @(#)    updated only  via the family  of header I/O routines  of which
C @(#)    the present routine is a member.  This design choice minimizes
C @(#)    the amount of memory that needs to be allocated for individual
C @(#)    Extension Headers,  providing  a larger degree  of flexibility
C @(#)    than a much simpler  'header(MAXEXT,MAXREC)' structure.   This
C @(#)    comes at a cost of extra computational overhead  in populating
C @(#)    new and, particularly, modifying existing multi-extension FITS
C @(#)    headers that grows with the total number of records within the
C @(#)    header.
C @(#)
C @(#)  Dependencies:	'buffers.inc' ;  cinit, clen, gdate ;  rjfphrec,
C @(#)    rjfpkwrd, rjfpkvali, rjfpkvall, rjfpkvale, rjfpkvals, rjfpkcom
C @(#)
C @(#)  Revision history:
C @(#)    Jul  8, 2009 [RAJ] -- Initial version of this routine.
C @(#)    Jul 10, 2009 [RAJ] -- Corrected flat-FITS signaling via 'iext'
C @(#)                          = -1 --> lflat=T; Updated documentation.
C @(#)    Sep  9, 2009 [RAJ] -- Fix of calls to 'rjfpkvals',  reflecting
C @(#)                          change in calling sequence ("l8" added);
C @(#)                          Fix of calls to 'rjfpkcom' due to change
C @(#)                          in calling sequence ("ltr" added).
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of maximum number of image extensions, header records,
C     and image axes...
      INCLUDE   'buffers.inc'

C --- Declaration of input/return variables...
      CHARACTER*80 fhdr(MAXREC)
      INTEGER	iext, nrec(MAXEXT), nfree(MAXEXT), idbg, ierr

C --- Declaration of local variables...
      CHARACTER	dbid*19, crec*80, vstr*80, rjf_version*80, cfmt*8
      INTEGER	i, j, irec, clen
      LOGICAL	lflat, l8, ltc

C --- Date/time stamp variables...
      CHARACTER*24 dstr, gdate

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C     Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfchdr): '
      CALL cinit(crec)

C --- Return for requested header unit number outside the valid range...
      IF (iext.LT.-1) ierr = 2201
      IF (iext.GE.MAXEXT) ierr = 2202
      IF (ierr.NE.0) GOTO 9

C --- Initialize local variables; counter 'irec' will keep track of the
C     number of records populated in the present Header Unit...
      irec  = 0
      lflat = .FALSE.
      IF (iext.EQ.-1) THEN
         lflat = .TRUE.
         iext = 0
      ENDIF
C     Setting l8 to .TRUE. forces a minimum length of 8 characters for
C     string-valued keywords (FITS standards earlier than v3.0) ...
      l8 = .FALSE.
C     Setting ltc to .TRUE. allows truncation of header keyword comments
      ltc = .FALSE.

C --- Initialize output variables...
      nrec(iext+1)  =  0
      nfree(iext+1) = 36


C --- Determine the starting element of the specified header unit; loop
C     must not be executed unless iext>0 ...
      i = 0
      DO j=1,iext
         i = i + nrec(j)
      ENDDO
      IF (MOD(i,36).NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*)
     +      dbid//'HUs are not aligned on 2880-b/36-record blocks.'
         ierr = 2203
         GOTO 9
      ENDIF
C     First element of the current Header Unit in the header array...
      i = i + 1

C     Test that sufficient storage remains for at least 36 records...
      IF (i.GT.(MAXREC-36)) THEN
         IF (idbg.GT.0) WRITE(*,*)
     +      dbid//'Insufficient space remaining in header array.'
         ierr = 2220
         GOTO 9
      ENDIF


C --- Populate the specified HU with the mandatory header keywords that
C     are appropriate for this header unit...
      IF (iext.EQ.0) THEN
C        We are initializing a Primary Header Unit...
         CALL rjfpkwrd (crec,'SIMPLE',idbg,ierr)
         CALL rjfpkvall(crec,.TRUE.,idbg,ierr)
         CALL rjfpkcom (crec,'File conforms to FITS standard',ltc,
     +                  idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2204
            GOTO 9
         ENDIF
      ELSE
C        We are initializing an Extension Header Unit...
         CALL rjfpkwrd (crec,'XTENSION',idbg,ierr)
         CALL rjfpkvals(crec,'IMAGE',.TRUE.,idbg,ierr)
         CALL rjfpkcom (crec,'IMAGE extension',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2205
            GOTO 9
         ENDIF
      ENDIF
      irec = irec + 1

      CALL rjfpkwrd (crec,'BITPIX',idbg,ierr)
      CALL rjfpkvali(crec,16,idbg,ierr)
      CALL rjfpkcom (crec,'Number of bits per data value',ltc,idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2206
         GOTO 9
      ENDIF
      irec = irec + 1

      CALL rjfpkwrd (crec,'NAXIS',idbg,ierr)
      CALL rjfpkvali(crec,0,idbg,ierr)
      CALL rjfpkcom (crec,'Number of data axes',ltc,idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2207
         GOTO 9
      ENDIF
      irec = irec + 1

      IF (iext.EQ.0) THEN
         CALL rjfpkwrd (crec,'EXTEND',idbg,ierr)
         IF (lflat) THEN
            CALL rjfpkvall(crec,.FALSE.,idbg,ierr)
	 ELSE
            CALL rjfpkvall(crec,.TRUE.,idbg,ierr)
         ENDIF
         CALL rjfpkcom (crec,
     +      'FITS file may contain standard extensions',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2208
            GOTO 9
         ENDIF
         irec = irec + 1

         IF (.NOT.lflat) THEN
            CALL rjfpkwrd (crec,'NEXTEND',idbg,ierr)
            CALL rjfpkvali(crec,0,idbg,ierr)
            CALL rjfpkcom (crec,'Number of standard extensions',ltc,
     +         idbg,ierr)
            CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
            IF (ierr.NE.0) THEN
               ierr = 2209
               GOTO 9
            ENDIF
            irec = irec + 1
         ENDIF

         vstr = rjf_version()
         CALL rjfpkwrd (crec,'ORIGIN',idbg,ierr)
         CALL rjfpkvals(crec,vstr,l8,idbg,ierr)
         CALL rjfpkcom (crec,'FITS file originator',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2210
            GOTO 9
         ENDIF
         irec = irec + 1
      ELSE
         CALL rjfpkwrd (crec,'PCOUNT',idbg,ierr)
         CALL rjfpkvali(crec,0,idbg,ierr)
         CALL rjfpkcom (crec,'Number of group parameters',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2211
            GOTO 9
         ENDIF
         irec = irec + 1

         CALL rjfpkwrd (crec,'GCOUNT',idbg,ierr)
         CALL rjfpkvali(crec,1,idbg,ierr)
         CALL rjfpkcom (crec,'Number of groups',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2212
            GOTO 9
         ENDIF
         irec = irec + 1

         CALL rjfpkwrd (crec,'EXTNAME',idbg,ierr)
         CALL rjfpkvals(crec,'image',l8,idbg,ierr)
         CALL rjfpkcom (crec,'Extension name',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2213
            GOTO 9
         ENDIF
         irec = irec + 1

         CALL rjfpkwrd (crec,'EXTVER',idbg,ierr)
         CALL rjfpkvali(crec,iext,idbg,ierr)
         CALL rjfpkcom (crec,'Extension version number',ltc,idbg,ierr)
         CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2214
            GOTO 9
         ENDIF
         irec = irec + 1
      ENDIF

      CALL cinit(cfmt)
      cfmt = 'F3.1'
      CALL rjfpkwrd (crec,'BZERO',idbg,ierr)
      CALL rjfpkvale(crec,0.0,cfmt,idbg,ierr)
      CALL rjfpkcom (crec,'Offset: value = BZERO + BSCALE*fits',ltc,
     +               idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2215
         GOTO 9
      ENDIF
      irec = irec + 1

      CALL cinit(cfmt)
      cfmt = 'F3.1'
      CALL rjfpkwrd (crec,'BSCALE',idbg,ierr)
      CALL rjfpkvale(crec,1.0,cfmt,idbg,ierr)
      CALL rjfpkcom (crec,'Scale factor: value = BZERO + BSCALE*fits',
     +               ltc,idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2216
         GOTO 9
      ENDIF
      irec = irec + 1

      dstr = gdate(1,0)
      CALL rjfpkwrd (crec,'DATE',idbg,ierr)
      CALL rjfpkvals(crec,dstr,l8,idbg,ierr)
      CALL rjfpkcom (crec,'Date this FITS file was written',ltc,
     +               idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2217
         GOTO 9
      ENDIF
      irec = irec + 1

      CALL rjfpkwrd (crec,'END',idbg,ierr)
      CALL rjfphrec ((i+irec),fhdr,crec,idbg,ierr)
      IF (ierr.NE.0) THEN
         ierr = 2218
         GOTO 9
      ENDIF
      irec = irec + 1

C --- Return number of free records within the current 36-record block..
      nfree(iext+1) = MOD(irec,36)

C --- Blank these free records within the current 36-record block; loop
C     must not be executed unless MOD(irec,36) > 0 ...
      DO j=1,MOD(irec,36)
         CALL cinit(crec)
         CALL rjfphrec((i+irec),fhdr,crec,idbg,ierr)
         IF (ierr.NE.0) THEN
            ierr = 2219
            GOTO 9
         ENDIF
         irec = irec + 1
      ENDDO

C --- Return the number of records populated in the requested header
C     unit number (including padding, if any)...
      nrec(iext+1)  = irec


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


