C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfgkvall (crec, lval, idbg, ierr)

C @(#)Subroutine: rjfgkvall          Author: R.A. Jansen -- Oct  5, 2009
C @(#)
C @(#)  Purpose:  Read a boolean-valued keyword value 'lval' from a FITS
C @(#)    header record 'crec' already known to contain a boolean-valued
C @(#)    keyword. If the keyword value was written in free format (this
C @(#)    is allowed in v3.0 of the FITS standard),  or non-compliant in
C @(#)    a straight-forward sense (ie, Y and N and variations thereof),
C @(#)    'crec' will be returned modified to fixed format upon exit. If
C @(#)    verbosity level 'idbg'>0,  additional messages may be printed.
C @(#)    Should any error be detected, error status 'ierr' will be non-
C @(#)    zero  upon exit.  Its value  (if >2000)  uniquely identifies a
C @(#)    known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfgkvall (crec, lval, idbg, ierr),   where  'crec'  is a
C @(#)    char string with a declared length of 80 characters and 'lval'
C @(#)    a boolean (LOGICAL) variable.  'idbg' and 'ierr' are integers.
C @(#)
C @(#)  Note: this task is meant to be run after 'rjfvkval' (when 'ityp'
C @(#)    has the return value of 2) and 'rjfgkwrd'.
C @(#)
C @(#)  Dependencies:	clen, cinit, cidx, lenc, ucase
C @(#)
C @(#)  Revision history:
C @(#)    Jul  1, 2009 [RAJ] -- Initial version ('gtfvbln ') of routine.
C @(#)    Oct  5, 2009 [RAJ] -- Major re-write of the code;  changed the
C @(#)                          calling sequence; updated documentation.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80
      INTEGER	idbg, ierr
      LOGICAL	lval

C --- Declaration of local variables...
      CHARACTER	ctmp*80, dbid*19, cc
      INTEGER	clen, lenc, cidx, i, j, k, ic

C     'ib','ie','id','is','iFF','iTT','iNN','iYY' are shorthand integer
C     equivalents in the ASCII collating sequence of the space, equals
C     sign, period, slash, and characters F, T, N and Y, respectively...
      INTEGER	ib, ie, id, is, iFF, iTT, iYY, iNN

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid = 'DEBUG(rjfgkvall): '

C --- Initialize the output variable...
      lval = .FALSE.

C --- Initialize the local variables...
      CALL cinit(ctmp)
      ib  = ICHAR(' ')
      ie  = ICHAR('=')
      id  = ICHAR('.')
      is  = ICHAR('/')
      iFF = ICHAR('F')
      iTT = ICHAR('T')
      iYY = ICHAR('Y')
      iNN = ICHAR('N')

C --- Test that byte 9 in 'crec' is indeed an equals sign (captures user
C     error of calling 'rjfgkvall' before 'rjfvkval')...
      cc = crec(9:9)
      ic = ICHAR(cc)
      IF (ic.NE.ie) THEN
C        Not an equals-sign, therefore no keyword value in this record!
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'header record does not contain a keyword value'
         ierr = 2701
         GOTO 9
      ENDIF

C --- Assign the nominal value portion of the header record to 'ctmp',
C     and convert alphabetical characters to upper case.  In v3.0 of the
C     FITS standard, free-format values are allowed to start anywhere
C     after character 10 ...
      ctmp(1:71) = crec(10:80)
      CALL ucase(ctmp,ctmp)


C     The following counters store: j = actual length (position of last
C     non-blank character) of 'ctmp', i = position of first non-blank
C     character in 'ctmp', k = position of the last character belonging
C     to the keyword value...
      j = clen(ctmp)
      i = lenc(ctmp)
      k = 0


C --- Test that the first non-blank character in the value portion of
C     the record is a 'T','F','Y','N' or '.' (captures the user error of
C     calling 'rjfgkvall' while the return value of 'rjfvkval' != 2) ...
      cc = ctmp(i:i)
      ic = ICHAR(cc)
      IF (ic.NE.iTT.AND.ic.NE.iFF
     +    .AND.ic.NE.iYY.AND.ic.NE.iNN.AND.ic.NE.id) THEN
C        Unexpected first non-blank char; keyword is not boolean-valued!
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value is not a '//
     +      'boolean (does not start with T or F [or N or Y])'
         ierr = 2702
         GOTO 9
      ENDIF


C --- Exhaustively test the valid (and otherwise recognized) options for
C     the subsequent non-blank characters. They should cover: T, F, TRUE,
C     FALSE, .TRUE., .FALSE., Y, N, YES, and NO ...
      cc = ctmp((i+1):(i+1))
      IF (ic.EQ.iTT) THEN
         k = i + 3
         IF (cc.NE.' '.AND.cc.NE.'/'.AND.ctmp(i:k).NE.'TRUE') THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted keyword '//
     +         'value starting with T, but neither T nor TRUE.'
            ierr = 2703
            GOTO 9
         ENDIF
         lval = .TRUE.
      ELSE IF (ic.EQ.iFF) THEN
         k = i + 4
         IF (cc.NE.' '.AND.cc.NE.'/'.AND.ctmp(i:k).NE.'FALSE') THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted keyword '//
     +         'value starting with F, but neither F nor FALSE.'
            ierr = 2704
            GOTO 9
         ENDIF
      ELSE IF (ic.EQ.iYY) THEN
         k = i + 2
         IF (cc.NE.' '.AND.cc.NE.'/'.AND.ctmp(i:k).NE.'YES') THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted non-compli'//
     +       'ant keyword value starting with Y, but neither Y nor YES.'
            ierr = 2705
            GOTO 9
         ENDIF
         IF (idbg.GT.1) WRITE(*,*) dbid//' non-compliant keyword '//
     +      'value interpreted as .TRUE.'
         lval = .TRUE.
      ELSE IF (ic.EQ.iNN) THEN
         k = i + 1
         IF (cc.NE.' '.AND.cc.NE.'/'.AND.ctmp(i:k).NE.'NO') THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted non-compli'//
     +       'ant keyword value starting with N, but neither N nor NO.'
            ierr = 2706
            GOTO 9
         ENDIF
         IF (idbg.GT.1) WRITE(*,*) dbid//' non-compliant keyword '//
     +      'value interpreted as .FALSE.'
      ELSE IF (ic.EQ.id) THEN
         k = i + 5
         IF (ctmp(i:k).NE.'.TRUE.'.AND.ctmp(i:(k+1)).NE.'.FALSE.') THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted non-compli'//
     +         'ant keyword value starting with a period, but neither'//
     +         '.TRUE. nor .FALSE.'
            ierr = 2707
            GOTO 9
         ENDIF
         IF (ctmp(i:k).EQ.'.TRUE.') THEN
            IF (idbg.GT.1) WRITE(*,*) dbid//' non-compliant keyword '//
     +         'value interpreted as .TRUE.'
            lval = .TRUE.
         ELSE IF (ctmp(i:(k+1)).EQ.'.FALSE.') THEN
            IF (idbg.GT.1) WRITE(*,*) dbid//' non-compliant keyword '//
     +         'value interpreted as .FALSE.'
            k = i + 6
         ENDIF
      ELSE
C        It should be impossible to arrive here...
         IF (idbg.GT.0) WRITE(*,*) dbid//' read failed impossibly'
         ierr = 2708
         GOTO 9
      ENDIF


C --- Return a possibly modified version of input header record 'crec',
C     with the keyword value (and comment) in fixed format...

C     First test whether the start position of any keyword comment needs
C     to be adjusted...
      j = cidx(crec((k+10):80),'/')
      j = j + (k+9)
      IF (k.GT.32) THEN
C        Shift keyword comment to the left...
         crec(32:(32+(80-j))) = crec(j:80)
         DO i = (32+(80-j)+1),80,1
            crec(i:i) = ' '
         ENDDO
      ELSE IF (j.LT.32) THEN
C        Shift keyword comment to the right...
         crec(32:80) = crec(j:(80-(32-j)))
         DO i = j,31,1
            crec(i:i) = ' '
         ENDDO
      ENDIF
C     Then write the keyword value itself in fixed format in byte 30...
      crec(11:31) = ' '
      crec(30:30) = 'F'
      IF (lval) crec(30:30) = 'T'



9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, lval=',lval,
     +   ', ierr=',ierr

      RETURN
      END


