C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfpkvalp (crec, rrval, rival, idbg, ierr)

C @(#)Subroutine: rjfpkvalp          Author: R.A. Jansen -- Sep  2, 2009
C @(#)
C @(#)  Purpose:  Write a  complex real-valued  keyword value  to a FITS
C @(#)    header record 'crec' for a keyword value of which the real and
C @(#)    imaginary parts are given by 'rrval' and 'rival' that are both
C @(#)    declared as variables of the default REAL data type.  For most
C @(#)    host OS  and  compiler implementations,  this corresponds to a
C @(#)    32-bit (REAL*4) floating point real. The keyword value will be
C @(#)    written using default formatting  as a comma-separated pair of
C @(#)    reals, enclosed in parentheses,  and left-justified on byte 11
C @(#)    in the header record.  If verbosity level 'idbg'>0, additional
C @(#)    messages may be printed.  Error status 'ierr' will be non-zero
C @(#)    upon exit, should any error condition be detected.   Its value
C @(#)    (if >2000) uniquely identifies a known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfpkvalp(crec, rrval, rival, idbg, ierr) ,  where 'crec'
C @(#)    is a character string with a declared length of  80 characters
C @(#)    and 'rrval' and 'rival' are variables of the default REAL data
C @(#)    type. 'idbg' and 'ierr' are both integers.
C @(#)
C @(#)  Note1: 32-bit floating point REALs generally represent at most 6
C @(#)    or 7 significant digits faithfully;  this should be taken into
C @(#)    account when using this task.
C @(#)  Note2: internally, scaling offset 1P is enforced for exponential
C @(#)    notation, such that a value of pi is written as  3.1415927E+00
C @(#)    instead of the Fortran default 0.31415927E+01.
C @(#)  Note3: this task is meant to write a complex keyword value  to a
C @(#)    header record that has been initialized with the FITS keyword;
C @(#)    use one of 'rjfmkval?' to modify an existing keyword value.
C @(#)
C @(#)  Dependencies:	clen, cinit
C @(#)
C @(#)  Revision history:
C @(#)    Sep  2, 2009 [RAJ] -- Initial version of this routine.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80
      REAL	rrval, rival
      INTEGER	idbg, ierr

C --- Declaration of local variables...
      CHARACTER dbid*19, ctmp*80, cval*80, cfmt*36, cc
      REAL	rrtmp, ritmp
      INTEGER	clen, ic, i, j

C     The following variables are shorthand integer equivalents in the
C     ASCII collating sequence of the digits 0 and 9, and the space, E,
C     +, -, and . , respectively...
      INTEGER	i0, i9, ib, iEE, ip, im, id

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfpkvalp): '

C --- Initialize local variables...
      CALL cinit(cval)
      CALL cinit(ctmp)
      CALL cinit(cfmt)
      i0  = ICHAR('0')
      i9  = ICHAR('9')
      ib  = ICHAR(' ')
      iEE = ICHAR('E')
      ip  = ICHAR('+')
      im  = ICHAR('-')
      id  = ICHAR('.')
      rrtmp = ABS(rrval)
      ritmp = ABS(rival)

C --- Build format specification to use for default formatting...
      cfmt(1:4) = '(A1,'
      i = clen(cfmt)
      IF (rrtmp.GT.1.0D-3.AND.rrtmp.LT.1.0D3) THEN
         IF (rrtmp.LT.1.0D-2) THEN
            cfmt(i+1:i+11) = 'SPF20.9,A1,'
         ELSE IF (rrtmp.LT.1.0D-1) THEN
            cfmt(i+1:i+11) = 'SPF20.8,A1,'
         ELSE IF (rrtmp.LT.1.0D1) THEN
            cfmt(i+1:i+11) = 'SPF20.7,A1,'
         ELSE IF (rrtmp.LT.1.0D2) THEN
            cfmt(i+1:i+11) = 'SPF20.6,A1,'
         ELSE
            cfmt(i+1:i+11) = 'SPF20.5,A1,'
         ENDIF
      ELSE IF (rrtmp.LT.1.0D100) THEN
         cfmt(i+1:i+15) = '1PSPE20.7E2,A1,'
      ELSE
         cfmt(i+1:i+15) = '1PSPE20.7E3,A1,'
      ENDIF
      i = clen(cfmt)

      IF (ritmp.GT.1.0D-3.AND.ritmp.LT.1.0D3) THEN
         IF (ritmp.LT.1.0D-2) THEN
            cfmt(i+1:i+11) = 'SPF20.9,A1)'
         ELSE IF (ritmp.LT.1.0D-1) THEN
            cfmt(i+1:i+11) = 'SPF20.8,A1)'
         ELSE IF (ritmp.LT.1.0D1) THEN
            cfmt(i+1:i+11) = 'SPF20.7,A1)'
         ELSE IF (ritmp.LT.1.0D2) THEN
            cfmt(i+1:i+11) = 'SPF20.6,A1)'
         ELSE
            cfmt(i+1:i+11) = 'SPF20.5,A1)'
         ENDIF
      ELSE IF (ritmp.LT.1.0D100) THEN
         cfmt(i+1:i+15) = '1PSPE20.7E2,A1)'
      ELSE
         cfmt(i+1:i+15) = '1PSPE20.7E3,A1)'
      ENDIF


C --- Try to write the real and imaginary parts of the complex keyword
C     value to temporary character string 'ctmp' using the format string
C     we just constructed...
      WRITE(ctmp,cfmt) '(',rrval,',',rival,')'


C --- Copy the non-blank characters to character string 'cval', left-
C     justified on byte 11, while testing on-the-fly that only digits,
C     decimal point, upper case E, plus, minus, comma and parentheses
C     were present...
      i = 10
      DO j=1,clen(ctmp)
         cc = ctmp(j:j)
         ic = ICHAR(cc)
         IF (ic.NE.ib) THEN
            IF ((ic.GE.i0.AND.ic.LE.i9).OR.ic.EQ.id.OR.ic.EQ.iEE
     +          .OR.ic.EQ.ip.OR.ic.EQ.im.OR.ic.EQ.40.OR.ic.EQ.41
     +          .OR.ic.EQ.44) THEN
               i = i + 1
               cval(i:i) = cc
            ELSE
C              Write appears to have failed (likely ***, or some such)..
               IF (idbg.GT.0) WRITE(*,*) dbid//
     +           'failed to write COMPLEX REAL keyword value to record.'
               ierr = 2441
               GOTO 9
            ENDIF
         ENDIF
      ENDDO
      i = i + 1
      cval(i:i) = ' '


C --- Try to read back the value written and compare with the original..
      CALL cinit(ctmp)
      ctmp = cval
      rrtmp = 9.99999E9
      ritmp = -9.99999E9
C     First, replace parentheses and comma by a blank..
      DO i=1,clen(ctmp)
         cc = ctmp(i:i)
         ic = ICHAR(cc)
         IF (ic.EQ.40.OR.ic.EQ.41.OR.ic.EQ.44) ctmp(i:i) = ' '
      ENDDO
C     Then try and read back real and imaginary parts as two reals...
      READ(ctmp,*,IOSTAT=ierr) rrtmp, ritmp
      IF (ierr.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +     'failed to read back COMPLEX REAL keyword value from record.'
         ierr = 2442
         GOTO 9
      ENDIF
      IF (1.0d-7.LT.(ABS(rrtmp-rrval)/rrval)) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//'real part of complex '//
     +      'keyword value read back from record does not match '//
     +      'input value.'
         ierr = 2443
         GOTO 9
      ENDIF
      IF (1.0d-7.LT.(ABS(ritmp-rival)/rival)) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//'imaginary part of complex '//
     +      'keyword value read back from record does not match '//
     +      'input value.'
         ierr = 2444
         GOTO 9
      ENDIF


C --- We appear to have successfully written the complex real keyword
C     value, so copy characters 11 and on to the output header record...
      j = clen(cval)
      crec(11:j) = cval(11:j)


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


