C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfpkwrd (crec, ckey, idbg, ierr)

C @(#)Subroutine: rjfpkwrd           Author: R.A. Jansen -- Jul 10, 2009
C @(#)
C @(#)  Purpose: Write a header keyword  'ckey'  to a FITS header record
C @(#)    'crec' after checking that it complies with the FITS standard,
C @(#)    where lowercase letters will be silently converted  into upper
C @(#)    case.  The standard requires a keyword to be a left-justified,
C @(#)    8-character, space-padded ASCII string with no embedded spaces
C @(#)    composed of uppercase letters (A-Z), numbers (0-9), hyphen (-)
C @(#)    and underscore (_). The first character is restricted to upper
C @(#)    case letter.   For indexed keyword names, the counter must not
C @(#)    have leading zeros. The keyword is followed on exit by a value
C @(#)    indicator ('= ') in position 9:10 of the record.  If verbosity
C @(#)    level 'idbg' > 0, additional messages may be printed.  Integer
C @(#)    error status 'ierr' will be non-zero on exit, should any error
C @(#)    condition be detected.   Its value (if >2000) uniquely identi-
C @(#)    fies a known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfpkwrd(crec, ckey, idbg, ierr), where 'crec' and 'ckey'
C @(#)    are strings with declared lengths of 80 and 8 characters, res-
C @(#)    pectively.  'idbg' and 'ierr' are both integers.
C @(#)
C @(#)  Note: this task is meant to initialize a new record by writing a
C @(#)    keyword to it.  Use 'rjfmkwrd' to modify a keyword in an exis-
C @(#)    ting record. This task is also not meant to initialize comment
C @(#)    records.  Use 'rjfpcomm' for that purpose.
C @(#)
C @(#)  Dependencies:	ucase, clen
C @(#)
C @(#)  Revision history:
C @(#)    Jul 10, 2009 [RAJ] -- Initial version of this routine.
C @(#)    Sep 14, 2009 [RAJ] -- Modified to adhere to naming conventions
C @(#)                          of more recent header routines.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80, ckey*8
      INTEGER	idbg, ierr

C --- Declaration of local variables...
      CHARACTER dbid*19, cc
      INTEGER	clen, ilen, ic, j

C     'iA','iZ','i0','i9','ih','iu' and 'ib' are shorthand integer equi-
C     valents in the ASCII collating sequence of uppercase letters A and
C     Z, numbers 0 and 9, hyphen, underscore, and space, respectively...
      INTEGER	iA, iZ, i0, i9, ih, iu, ib

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfpkwrd): '
 
C --- Return for a blank input header keyword...
      ilen = clen(ckey)
      IF (ilen.EQ.0) THEN
         IF (idbg.GT.0) WRITE(*,*)dbid//'invalid blank keyword on input'
         ierr = 2321
         GOTO 9
      ENDIF

C --- Initialize the local variables...
      iA = ICHAR('A')
      iZ = ICHAR('Z')
      i0 = ICHAR('0')
      i9 = ICHAR('9')
      ih = ICHAR('-')
      iu = ICHAR('_')
      ib = ICHAR(' ')

C --- Initialize the output FITS record with an uppercase version of the
C     input keyword; update the keyword string for further testing...
      CALL ucase(ckey,crec)
      ckey(1:8) = crec(1:8)
      crec(9:9) = '='

C --- Test whether the 1st character in the keyword starts with an upper
C     case letter...
      cc = ckey(1:1)
      ic = ICHAR(cc)
      IF (ic.LT.iA.OR.ic.GT.iZ) THEN
C        First character of keyword is not a capital letter...
         IF (idbg.GT.0) WRITE(*,*)
     +      dbid//'keyword does not start with a letter'
         ierr = 2322
         GOTO 9
      ENDIF

C --- Test that characters 2--8 comply with the standard...
      DO j=2,ilen,1
         cc = ckey(j:j)
         ic = ICHAR(cc)
         IF ( (ic.GE.iA.AND.ic.LE.iZ) .OR. (ic.GE.i0.AND.ic.LE.i9)
     +       .OR. ic.EQ.ih .OR. ic.EQ.iu ) THEN
C           character is compliant with the FITS standard...
            CONTINUE
         ELSE IF (ic.EQ.ib) THEN
C           keyword has invalid embedded blank character(s)...
            IF (idbg.GT.0) WRITE(*,*) dbid//'invalid keyword with '//
     +         ' embedded blank(s)'
            ierr = 2323
            GOTO 9
         ELSE
C           keyword contains invalid non-printable character(s)...
            IF (idbg.GT.0) WRITE(*,*) dbid//'invalid keyword with '//
     +         'embedded non-printable character(s)'   
            ierr = 2324
            GOTO 9
         ENDIF
      ENDDO

C --- Test that if we're dealing with an indexed keyword, there are no
C     leading zeros...
      cc = ckey(ilen:ilen)
      ic = ICHAR(cc)
      IF (ilen.GT.2.AND.ic.GE.i0.AND.ic.LE.i9) THEN
C        This looks like it might be an indexed keyword; find the last
C        non-digit (we know that the first character is not a digit)...
         DO j=ilen-1,1,-1
            cc = ckey(j:j)
            ic = ICHAR(cc)
            IF (.NOT.(ic.GE.i0.AND.ic.LE.i9)) GOTO 1
         ENDDO
1        CONTINUE
         IF ((ilen-1-j).GT.0) THEN
C           There were additional digits; test for a leading zero...
            cc = ckey((j+1):(j+1))
            ic = ICHAR(cc)
            IF (ic.EQ.i0) THEN
C              Indexed keyword has a leading zero: not OK.
               IF (idbg.GT.1) WRITE(*,*) dbid//' invalid keyword: '//
     +            'leading zero(s) in indexed keyword.'
               ierr = 2325
               GOTO 9
            ELSE
C              Index number starts with a non-zero digit: OK.
               ierr = 0
               GOTO 9
            ENDIF
         ENDIF
      ENDIF

C --- Remove the value indicator in case keyword matches one of the
C     reserved comment keywords: "HISTORY" and "COMMENT" (really should
C     have used routine 'rjfpcomm' for comment records)...
      IF (ckey.EQ.'HISTORY '.OR.ckey.EQ.'COMMENT ') crec(9:9) = ' '


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


