C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE wcfitsi(filename,pix,ncol,nrow,update,idbg,ierr)

C @(#)Subroutine: wcfitsi            Author: R.A. Jansen -- Apr 30, 2013
C @(#)
C @(#)  Purpose: Write a 2-D pixel buffer 'pix' to a new FITS file or to
C @(#)    the specified extension in an existing FITS file  (if 'update'
C @(#)    is .TRUE.), given the size of that buffer ('ncol'x'nrow'). One
C @(#)    may specify an image extension as part of the file name by ap-
C @(#)    pending its number within square braces (eg., "img.fits[2]" to
C @(#)    specify the 2nd extension; [0] denotes the primary header in a
C @(#)    MEF file).  When verbosity level 'idbg'>0, additional messages
C @(#)    may be printed.  Should any error condition be detected, error
C @(#)    status 'ierr' will be non-zero upon exit.
C @(#)
C @(#)  Calling sequence: 
C @(#)    CALL wcfitsi (filename,pix,ncol,nrow,update,idbg,ierr) , where
C @(#)    'filename' is a character string of arbitrary declared length,
C @(#)    'pix' is an integer valued array of size 'ncol'x'nrow' pixels,
C @(#)    (with both 'ncol' and 'nrow' integers),  'update' is a boolean
C @(#)    switch, and both 'idbg' and 'ierr' are integers.
C @(#)
C @(#)  Dependencies: several routines from the CFITSIO v3.0 I/O library
C @(#)    ('libcfitsio.a'; W.D. Pence), called using their f77 wrappers;
C @(#)    several routines from my 'rjgen' library:  cidx, cinit, cldx,
C @(#)            clen, ifmt, error, wsizstr
C @(#)
C @(#)  Note: compilation requires linking against both 'libcfitsio' and
C @(#)    against 'libm' (the standard math library).
C @(#)
C @(#)  Revision history:
C @(#)    Apr 30 2013 [RAJ] -- Wrote outline of this routine.
C @(#)    May 13 2013 [RAJ] -- Implemented updating existing FITS file.
C @(#) -----------------------------------------------------------------

      IMPLICIT NONE

C --- Learn about default devices...
      INCLUDE	'devices.inc'

C --- Learn about maximum image dimension, maximum axis lengths...
      INCLUDE	'buffers.inc'

C --- Declaration of input/return variables...
      CHARACTER*(*) filename
      INTEGER	pix(MAXSIZ2*MAXSIZ1), ncol, nrow, idbg, ierr
      LOGICAL	update

C --- File I/O variables...
      INTEGER	exists, readwrite, blocksize

C --- Memory I/O variables...
      INTEGER	tmppix(ncol*nrow)
      INTEGER	group, pix1, ihdu, nhdu, bpp, naxes, axlen(2)

C --- External functions...
      INTEGER	cidx, cldx, clen, ifmt

C --- Other program variables and loop counters...
      CHARACTER	ctmp*80, cext*8, dbid*19, fmt*32, cq*1
      INTEGER	mlen, npix, i, j, k, m

C     ------------------------------------------------------------------


C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Set subroutine/function identifier...
      dbid = 'DEBUG(wcfitsi):    '

C --- Initialize local variables...
      CALL cinit(ctmp)
      CALL cinit(cext)
      CALL cinit(fmt)
      mlen     = clen(filename)
      bpp      = 16
      naxes    = 2
      axlen(1) = ncol
      axlen(2) = nrow
      cq       = CHAR(39)


C --- Check whether the image name includes an extension specifier...
      i = cidx(filename,'[')
      j = cldx(filename,']')
      IF (i.GT.0.AND.j.GT.i) THEN
C        The filename includes an extension specifier...
         k = j - i - 1
         m = j - i + 1
         ctmp(1:m) = filename(i:j)
         IF (k.GT.3) THEN
            WRITE(STDOUT,'(A)') dbid//' cext='//ctmp(1:m)
            CALL error('wcfitsi','malformatted extension specifier')
            ierr = 1
            RETURN
         ENDIF
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' cext='//ctmp(1:m)
         cext = '  '//filename(i+1:j-1)//' '
         READ (cext,'(I3)',ERR=9) ihdu
      ELSE
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' cext=""'
         ihdu = 0
      ENDIF
      IF (idbg.GT.0) THEN
         k = ifmt(ihdu)
         WRITE(fmt,'((A),I1,A1)') '((A),I',k,')'
         IF (update) THEN
            WRITE(STDOUT,fmt) dbid//'updating HDU = ',ihdu
         ELSE
            WRITE(STDOUT,fmt) dbid//'creating HDU = ',ihdu
         ENDIF
      ENDIF


C --- Check that the file actually exists when updating and does not
C     already exist when creating a new FITS file...
      IF (idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTEXIST'
      CALL ftexist (filename(1:mlen), exists, ierr)
      IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
      IF (ierr.NE.0) THEN
         ctmp = 'couldnot verify existence of "'//filename(1:mlen)//'"'
         CALL error('wcfitsi',ctmp)
         ierr = 2
      ELSE IF (update.AND.exists.NE.1) THEN
         ctmp = 'cannot update "'//filename(1:mlen)//'": file not found'
         CALL error('wcfitsi',ctmp)
         ierr = 3
      ELSE IF (.NOT.update.AND.exists.EQ.1) THEN
         ctmp = 'cannot create "'//filename(1:mlen)//'": file exists'
         CALL error('wcfitsi',ctmp)
         ierr = 4
      ENDIF
      IF (ierr.NE.0) RETURN


C --- Copy the full pixel buffer to a temporary linear buffer of just
C     the required size (so we don't have to worry about the internal
C     layout in memory of the relevant FITS array) ...
      IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' copying pix --> tmppix'
      DO j=1,nrow
         DO i=1,ncol
            tmppix(i+(j-1)*ncol) = pix(i+(j-1)*ncol)
         ENDDO
      ENDDO


      IF (.NOT.update) THEN

C        Open and initialize a new, empty, FITS file...
         blocksize = 1
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTINIT'
         CALL ftinit (STDOFIL, filename, blocksize, ierr)
         IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
         IF (ierr.NE.0) THEN
            ctmp = 'creating new FITS file "'//filename(1:mlen)//'"'
            CALL error('wcfitsi',ctmp)
            ierr = 5
            RETURN
         ENDIF

C        Write the required header keywords to the FITS file...
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTPHPS'
         CALL ftphps (STDOFIL, bpp, naxes, axlen, ierr)
         IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
         IF (ierr.NE.0) THEN
            ctmp = 'writing minimal header to "'//filename(1:mlen)//'"'
            CALL error('wcfitsi',ctmp)
            ierr = 6
            RETURN
         ENDIF

C        Append a few additional header records...
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTPKYS'
         CALL ftpkys (STDOFIL, 'filename', filename(1:mlen),
     +                'Name of FITS file', ierr)
         IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTPKYS'
         CALL ftpkys (STDOFIL, 'imagetyp', 'MASK    ', 'IRAF type key',
     +                ierr)
         IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
         IF (ierr.NE.0) THEN
            ctmp = 'writing other keywords to "'//filename(1:mlen)//'"'
            CALL error('wcfitsi',ctmp)
            ierr = 7
            RETURN
         ENDIF

      ELSE

C        Open the existing FITS file in readwrite access mode and move
C        to either specified extension or the primary HDU ...
         blocksize = 1
         readwrite = 1
         IF (ihdu.EQ.0) THEN
            IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTOPEN'
            CALL ftopen (STDOFIL, filename, readwrite, blocksize, ierr)
         ELSE
            IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTNOPEN'
            CALL ftnopn (STDOFIL, filename, readwrite, ierr)
         ENDIF
         IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
         IF (ierr.NE.0) THEN
            ctmp = 'updating FITS file "'//filename(1:mlen)//'"'
            CALL error('wcfitsi',ctmp)
            ierr = 8
            RETURN
         ENDIF

C        Check which HDU in the file was actually opened...
         IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTGHDN'
         CALL ftghdn (STDOFIL, nhdu)
         IF (nhdu.NE.(ihdu+1)) THEN
            k = ifmt(nhdu)
            WRITE(fmt,'((A),I1,(A))') '((A),I',k,',A1)'
            WRITE(ctmp,fmt) 'could not move to specified HDU (nhdu=',
     +                      nhdu,')'
            CALL error('wcfitsi',ctmp)
            ierr = 9
            RETURN
         ENDIF

      ENDIF


C --- Write the temporary linear buffer to the FITS file...
      group = 1
      pix1  = 1
      npix  = ncol*nrow
      IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTPPRJ'
      CALL ftpprj (STDOFIL, group, pix1, npix, tmppix, ierr)
      IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
      IF (ierr.NE.0) THEN
         ctmp = 'writing pixel data to "'//filename(1:mlen)//'"'
         CALL error('wcfitsi',ctmp)
         ierr = 10
         RETURN
      ENDIF


C --- Append or update the DATE keyword in the current header unit...
      IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTPDAT'
      CALL ftpdat (STDOFIL, ierr)
      IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
      IF (ierr.NE.0) THEN
         ctmp = 'writing DATE keywords to "'//filename(1:mlen)//'"'
         CALL error('wcfitsi',ctmp)
         ierr = 11
         RETURN
      ENDIF


C --- Close the opened FITS file...
      IF(idbg.GT.1) WRITE(STDOUT,'(A)') dbid//' calling FTCLOS'
      CALL ftclos (STDOFIL, ierr)
      IF (ierr.NE.0.AND.idbg.GT.1) CALL ftrprt ('STDOUT', ierr)
      IF (ierr.NE.0) THEN
         ctmp = 'closing "'//filename(1:mlen)//'"'
         CALL error('wcfitsi',ctmp)
         ierr = 12
         RETURN
      ENDIF

      RETURN


C --- Report error condition from READ statement...
9     CALL cinit(ctmp)
      ctmp = 'reading image extension number from input string'
      CALL error('wcfitsi',ctmp)
      ierr = 1
      RETURN

      END


